#pragma once

#include "glbdef.h"
#include "pos.h"
#include "mob.h"
#include "item.h"

class ATLAS;

struct PT2
{
    int	x, y;
    int data;
};

struct ROOM_DEF
{
    PT2 size;
    const SQUARE_NAMES *squarelist;
    
    // Per square lists
    const PT2 *squareflaglist;
    const PT2 *itemlist;
    const PT2 *moblist;
    const PT2 *itemtypelist;
    const PT2 *moblevellist;
    const PT2 *intrinsiclist;
    const PT2 *signpostlist;
    const PT2 *linklist;

    // Global defines
    // I really should have had map flags!!!
    int		level, placeholder;
    u8		rarity;
    u8		globalmapflags;
    u8		onlyonce;
    u8		allowstairs;
    u8		allowgeneration;
    u8		allowdig;
    u8		allowitems;
    u8		allowtel;
    int		x, y;
    const char *quest;
    const char *north, *east, *south, *west, *up, *down;
    const char	*name;		// Specific room name from file name.
};

struct ROOM_DEF_NC
{
    ~ROOM_DEF_NC()
    {
	delete [] squarelist;
	delete [] squareflaglist;
	delete [] itemlist;
	delete [] moblist;
	delete [] itemtypelist;
	delete [] moblevellist;
	delete [] intrinsiclist;
	delete [] signpostlist;
	delete [] linklist;
	// We do leak our names...
    }
    PT2 size;
    SQUARE_NAMES *squarelist;
    
    // Per square lists
    PT2 *squareflaglist;
    PT2 *itemlist;
    PT2 *moblist;
    PT2 *itemtypelist;
    PT2 *moblevellist;
    PT2 *intrinsiclist;
    PT2 *signpostlist;
    PT2 *linklist;

    // Global defines
    int		level, placeholder;
    u8		rarity;
    u8		globalmapflags;
    u8		onlyonce;
    u8		allowstairs;
    u8		allowgeneration;
    u8		allowdig;
    u8		allowitems;
    u8		allowtel;
    int		x, y;
    int		width() const { return size.x; }
    int		height() const { return size.y; }
    const char *quest;
    const char *north, *east, *south, *west, *up, *down;
    const char *name;		// Specific room name from file name.
};
class MAP
{
public:
    MAP(const ROOM_DEF &roomdef);
    ~MAP();

    void		 resetAll();

    const ROOM_DEF	&defn() const { return *myDef; }

    int			 level() const { return myLevel; }
    void		 incLevel() { myLevel++; }

    bool		 isSurface() const { return mySurface; }

    POS			 findRandomSquare(SQUARE_NAMES square) const;
    POS			 findRandomItem(ITEM_NAMES item) const;
    POS			 findRandomPassable() const;
    void		 heartbeat(int phase);
    void		 clearMessages();
    MOB			*findMobByType(MOB_NAMES mob) const;

    template <typename FILTER>	 
    POS			 findRandomPos(FILTER &filter) const;
    template <typename FILTER>	 
    POS			 findRandomPos(POS oldpos, FILTER &filter, int *chance) const;

    int			 countTopItems(ITEM_NAMES item) const;

    class iterator
    {
    public:
	iterator();
	iterator(const MAP *map);
	void	advance();
	POS	pos() const;
	bool	done() const;
    protected:
	int	x, y;
	const MAP 	*myMap;
    };

    iterator	begin() const;

    void		 makePosValid(POS &pos);

    void		 registerMob(MOB *mob);
    void		 deregisterMob(MOB *mob);
    void		 updateMobCache(MOB *mob, POS oldpos, POS newpos);

    void		 registerItem(ITEM *item);
    void		 deregisterItem(ITEM *item);
    void		 updateItemCache(ITEM *item, POS oldpos, POS newpos);

    static void		 render(POS center, DIR dir, MOB *avatar);

    void		 clear(SQUARE_NAMES square);
    void		 resetFlag(FLAG_NAMES flag, bool val = false);

    ATLAS		*atlas() const { return myAtlas; }
    const char		*name() const { return myName.buffer(); }
    int			 id() const { return myMapId; }
    void		 setAtlas(ATLAS *atlas, int mapid);

    int			 width() const { return myWidth; }
    int			 height() const { return myHeight; }

    // Builds width*height rgb data of overview
    u8			*buildMiniMap() const;

    // Ensures every thing that should have a link has a link.
    // Rules are:
    // Levers have only one link on them, and that links at least one place
    // Tele has one link on them, and that links only one place
    // Spams to console failed links
    int 		 validateLinks(bool quiet) const;

protected:
    SQUARE_NAMES	 getTile(int x, int y) const;
    void		 setTile(int x, int y, SQUARE_NAMES square);
    ITEM		*getItem(int x, int y) const;
    void		 setItem(int x, int y, ITEM *item);
    int 		 getItemList(int x, int y, ITEMLIST &list) const;
    MOB			*getMob(int x, int y) const;
    void		 setMob(int x, int y, MOB *mob);
    bool		 getFlag(FLAG_NAMES flag, int x, int y) const;
    void		 setFlag(FLAG_NAMES flag, int x, int y, bool val);

public:
    // Find links from this positionos
    void		 getLinks(int x, int y, PTRLIST<int> &links) const;
    // Finds the positions that have this link.
    void		 findLinks(int link, POSLIST &plist) const;

protected:
    // Searches the edge for a map link, id 100.
    // When found returns its edge coordinate.
    // -1 is used for free edge, 0 or 1 for fixed.
    int			 findMapLink(int x, int y) const;

    bool		 validPos(int x, int y) const;
    POS			 buildPos(int x, int y) const;

    SQUARE_NAMES	*mySquares;
    ITEM	 	**myItemCache;
    MOB			**myMobCache;
    u8			*myFlags;
    
    int			 myWidth, myHeight;

    ITEMLIST		 myItemList;
    MOBLIST		 myMobList;

    ATLAS		*myAtlas;
    int			 myMapId;
    int			 myLevel;
    bool		 mySurface;
    const ROOM_DEF	*myDef;

    BUF			 myName;

    friend class POS;
    friend class ATLAS;
    friend class iterator;
};

template <typename FILTER>
POS
MAP::findRandomPos(FILTER &filter) const
{
    POS		pos = POS::invalidPos();

    pos = findRandomPos(pos, filter, 0);
    return pos;
}

template <typename FILTER>
POS
MAP::findRandomPos(POS pos, FILTER &filter, int *chance) const
{
    int		x, y;
    int		count = 0, choice;
    POS		p = POS::invalidPos();

    FORALL_XY(x, y)
    {
	p = buildPos(x, y);
	if (filter(p))
	{
	    count++;
	}
    }

    if (!count)
	return pos;

    if (chance)
    {
	choice = rand_choice(count+*chance);
	*chance += count;
    }
    else
	choice = rand_choice(count);

    if (choice > count)
	return pos;

    FORALL_XY(x, y)
    {
	p = buildPos(x, y);
	if (filter(p))
	{
	    if (!choice)
	    {
		return p;
	    }
	    choice--;
	}
    }
    return POS::invalidPos();
}
